from __future__ import absolute_import
import logging
import os
import urllib2

import yaml

from cassette.http_response import MockedHTTPResponse
from cassette.addinfourl import MockedAddInfoURL

old_urlopen = urllib2.urlopen
log = logging.getLogger("cassette")


class CassetteName(unicode):

    """
    A CassetteName represents an unique way to retrieve the cassette
    from the library.
    """

    @classmethod
    def from_urlopen_url(cls, url, data):
        """Create an object from a urlopen url."""
        req = urllib2.Request(url, data)
        return cls.from_urlopen_request(req)

    @classmethod
    def from_urlopen_request(cls, request, data=None):
        """Create an object from a urlopen request."""

        # From Python source code, urllib2.py:382
        if data is not None:
            request.add_data(data)

        data = request.get_data()
        if not data:
            data = ""

        template_vars = dict(
            method=request.get_method(),
            url=request.get_full_url(),
            data=data
        )

        name = "urlopen:{method} {url} {data}".format(**template_vars)
        name = name.strip()
        return name

    @classmethod
    def from_httplib_connection(cls, host, port, method, url, body):
        """Create an object from an httplib request."""

        name = "httplib:{method} {host}:{port}{url} {body}".format(**locals())
        name = name.strip()
        return name


class CassetteLibrary(object):

    def __init__(self, filename):
        """Create a CassetteLibrary object.

        :param str filename: filename to file holding fake response
        """

        self.filename = os.path.abspath(filename)
        self.data = self.load_file()

    def add_response(self, cassette_name, response):
        """Add a new response to the mocked response.

        :param str cassette_name:
        :param response:
        """

        if not cassette_name:
            raise TypeError("No cassette name provided.")

        # Choose the class based on the response class
        if hasattr(response, "getheaders"):
            # It's an httplib.HTTPResponse
            mock_response_class = MockedHTTPResponse

        else:
            mock_response_class = MockedAddInfoURL

        mocked = mock_response_class.from_response(response)
        self.data[cassette_name] = mocked
        # TODO: would be probably more efficient to write the file only
        # on certain occasion.
        self.write_to_file()

        return mocked

    def write_to_file(self):
        """Write mocked responses to file."""

        data = {k: v.to_dict() for k, v in self.data.items()}
        with open(self.filename, "w+") as f:
            yaml.dump(data, f)

    def load_file(self):
        """Load MockedResponses from YAML file."""

        data = {}

        if not os.path.exists(self.filename):
            log.info("File '%s' does not exist." % self.filename)
            return data

        with open(self.filename) as f:
            content = yaml.load(f)

        for k, v in content.items():
            if k.startswith("urlopen:"):
                mock_response_class = MockedAddInfoURL
            else:
                mock_response_class = MockedHTTPResponse

            data[k] = mock_response_class.from_dict(v)

        return data

    def cassette_name_for_urlopen(self, url_or_request, data):
        """Create a cassette name from a urlopen request."""

        if isinstance(url_or_request, basestring):
            name = CassetteName.from_urlopen_url(url_or_request, data)
        else:
            # url is a urllib2.Request object
            name = CassetteName.from_urlopen_request(url_or_request, data)

        return name

    def cassette_name_for_httplib_connection(self, host, port, method,
                                             url, body):
        """Create a cassette name from an httplib request."""
        return CassetteName.from_httplib_connection(
            host, port, method, url, body)

    def _had_response(self):
        """Mark that the library already the response.

        This is for testing purposes (it's complicated to patch the unpatched
        version of urllib2/httplib).
        """
        pass

    def __contains__(self, cassette_name):
        contains = cassette_name in self.data

        if contains:
            self._had_response()  # For testing purposes
            log.info("Library has '%s'" % cassette_name)

        else:
            log.warning("Library does not have '%s'" % cassette_name)

        return contains

    def __getitem__(self, cassette_name):
        return self.data[cassette_name]



from __future__ import absolute_import
import urllib2
import httplib

from cassette.urlopen import CassetteURLOpen
from cassette.http_connection import CassetteHTTPConnection, CassetteHTTPSConnection

unpatched_urlopen = urllib2.urlopen
unpatched_HTTPConnection = httplib.HTTPConnection
unpatched_HTTPSConnection = httplib.HTTPSConnection


def patch(cassette_library):
    """Replace standard library."""

    urllib2.urlopen = CassetteURLOpen(cassette_library)

    # Taken from vcrpy

    CassetteHTTPConnection._cassette_library = cassette_library
    CassetteHTTPSConnection._cassette_library = cassette_library

    httplib.HTTPConnection = CassetteHTTPConnection
    httplib.HTTP._connection_class = CassetteHTTPConnection
    httplib.HTTPSConnection = CassetteHTTPSConnection
    httplib.HTTPS._connection_class = CassetteHTTPSConnection


def unpatch():
    """Unpatch standard library."""

    urllib2.urlopen = unpatched_urlopen

    # Taken from vcrpy

    httplib.HTTPConnection = unpatched_HTTPConnection
    httplib.HTTP._connection_class = unpatched_HTTPConnection
    httplib.HTTPSConnection = unpatched_HTTPSConnection
    httplib.HTTPS._connection_class = unpatched_HTTPSConnection



# -*- coding: utf8 -*-
"""
Test the cassette behavior.
"""

import httplib
import json
import os
import unittest
import urllib
import urllib2

import mock

import cassette
from cassette.cassette_library import CassetteLibrary


TEMPORARY_RESPONSES_FILENAME = "./cassette/tests/data/responses.temp.yaml"
RESPONSES_FILENAME = "./cassette/tests/data/responses.yaml"
TEST_HOST = "http://127.0.0.1:5000/"
TEST_URL = "http://127.0.0.1:5000/index"
TEST_URL_HTTPS = "https://httpbin.org/ip"
TEST_URL_REDIRECT = "http://127.0.0.1:5000/will_redirect"


# Taken from requests
def to_key_val_list(value):
    """Take an object and test to see if it can be represented as a
    dictionary. If it can be, return a list of tuples, e.g.,

    ::

    >>> to_key_val_list([('key', 'val')])
    [('key', 'val')]
    >>> to_key_val_list({'key': 'val'})
    [('key', 'val')]
    >>> to_key_val_list('string')
    Traceback (most recent call last):
        ...
    ValueError: cannot encode objects that are not 2-tuples
    """

    if value is None:
        return None

    if isinstance(value, (str, bytes, bool, int)):
        raise ValueError('cannot encode objects that are not 2-tuples')

    if isinstance(value, dict):
        value = value.items()

    return list(value)


def _encode_params(data):
    """Encode parameters in a piece of data.

    Will successfully encode parameters when passed as a dict or a list of
    2-tuples. Order is retained if data is a list of 2-tuples but abritrary
    if parameters are supplied as a dict.
    """

    if isinstance(data, (str, bytes)):
        return data
    elif hasattr(data, 'read'):
        return data
    elif hasattr(data, '__iter__'):
        result = []
        for k, vs in to_key_val_list(data):
            if isinstance(vs, basestring) or not hasattr(vs, '__iter__'):
                vs = [vs]
            for v in vs:
                if v is not None:
                    result.append(
                        (k.encode('utf-8') if isinstance(k, str) else k,
                         v.encode('utf-8') if isinstance(v, str) else v))
        return urllib.urlencode(result, doseq=True)


def url_for(endpoint):
    """Return full URL for endpoint."""
    return TEST_HOST + endpoint

#
# Testing the whole flow with a temporary response file.
#


class TestCassette(unittest.TestCase):

    def setUp(self):

        # This is a dummy method that we use to check if cassette had
        # the response.
        patcher = mock.patch.object(CassetteLibrary, "_had_response")
        self.had_response = patcher.start()
        self.addCleanup(patcher.stop)

        if os.path.exists(TEMPORARY_RESPONSES_FILENAME):
            os.remove(TEMPORARY_RESPONSES_FILENAME)

    def tearDown(self):

        if os.path.exists(TEMPORARY_RESPONSES_FILENAME):
            os.remove(TEMPORARY_RESPONSES_FILENAME)

    def check_urllib2_flow(self, url, expected_content=None,
                           allow_incomplete_match=False,
                           data=None):
        """Verify the urllib2 flow."""

        if not url.startswith("http"):
            url = url_for(url)

        # First run
        with cassette.play(TEMPORARY_RESPONSES_FILENAME):
            r = urllib2.urlopen(url, data)  # 1st run

        self.assertEqual(self.had_response.called, False)
        if expected_content:
            if allow_incomplete_match:
                self.assertIn(expected_content, r.read())
            else:
                self.assertEqual(r.read(), expected_content)

        self.had_response.reset_mock()

        # Second run
        with cassette.play(TEMPORARY_RESPONSES_FILENAME):
            r = urllib2.urlopen(url, data)  # 2nd run

        self.assertEqual(self.had_response.called, True)
        if expected_content:
            if allow_incomplete_match:
                self.assertIn(expected_content, r.read())
            else:
                self.assertEqual(r.read(), expected_content)

        if r.headers["Content-Type"] == "application/json":
            try:
                r.json = json.loads(r.read())
            except ValueError:
                pass

        return r

    def test_flow(self):
        """Verify that cassette works when using urllib2.urlopen."""
        self.check_urllib2_flow(TEST_URL, "hello world")

    def test_flow_redirected(self):
        """Verify that cassette works when redirected."""
        self.check_urllib2_flow(TEST_URL_REDIRECT, "hello world redirected")

    def test_flow_httplib(self):
        """Verify that cassette works when using httplib directly."""

        # First run
        with cassette.play(TEMPORARY_RESPONSES_FILENAME):
            conn = httplib.HTTPConnection("127.0.0.1", 5000)
            conn.request("GET", "/index")
            r = conn.getresponse()

        self.assertEqual(r.status, 200)
        self.assertEqual(r.reason, "OK")
        self.assertEqual(r.read(), "hello world")
        self.assertEqual(self.had_response.called, False)

        self.had_response.reset_mock()

        # Second run
        with cassette.play(TEMPORARY_RESPONSES_FILENAME):
            conn = httplib.HTTPConnection("127.0.0.1", 5000)
            conn.request("GET", "/index")
            r = conn.getresponse()

        self.assertEqual(r.status, 200)
        self.assertEqual(r.reason, "OK")
        self.assertEqual(r.read(), "hello world")
        self.assertEqual(self.had_response.called, True)

    def test_flow_https(self):
        """Verify the cassette behavior for HTTPS."""
        self.check_urllib2_flow(TEST_URL_HTTPS, "origin",
                                allow_incomplete_match=True)

    def test_flow_manual_context(self):
        """Verify the cassette behavior when setting up the context."""

        # First run
        cassette.insert(TEMPORARY_RESPONSES_FILENAME)
        r = urllib2.urlopen(TEST_URL)
        cassette.eject()

        self.assertEqual(self.had_response.called, False)
        self.assertEqual(r.read(), "hello world")

        self.had_response.reset_mock()

        # Second run
        cassette.insert(TEMPORARY_RESPONSES_FILENAME)
        r = urllib2.urlopen(TEST_URL)
        cassette.eject()

        self.assertEqual(self.had_response.called, True)
        self.assertEqual(r.read(), "hello world")

    def test_flow_get_non_ascii_page(self):
        """Verify that cassette can store a non-ascii page."""
        self.check_urllib2_flow(
            url="http://127.0.0.1:5000/non-ascii-content",
            expected_content=u"Le Mexicain l'avait achetée en viager "
            u"à un procureur à la retraite. Après trois mois, "
            u"l'accident bête. Une affaire.")

    def test_flow_get_with_array_args(self):
        """Verify that cassette can store array args."""

        param = {
            "dict1": {"dict2": "dict3", "dict4": "dict5"},
            "array": ["item1", "item2"],
            "int": 1,
            "param": "1",
        }

        url = "/get?"
        url += _encode_params(param)
        r = self.check_urllib2_flow(url=url)
        self.assertEqual(r.json["args"]["param"], "1")

#
# Verify that cassette can read from an existing file.
#


class TestCassetteFile(unittest.TestCase):

    def setUp(self):

        # This is a dummy method that we use to check if cassette had
        # the response.
        patcher = mock.patch.object(CassetteLibrary, "_had_response")
        self.had_response = patcher.start()
        self.addCleanup(patcher.stop)

    def check_read_from_file_flow(self, url, expected_content,
                                  allow_incomplete_match=False):
        """Verify the flow when reading from an existing file."""

        with cassette.play(RESPONSES_FILENAME):
            r = urllib2.urlopen(url)

        self.assertEqual(self.had_response.called, True)
        if allow_incomplete_match:
            self.assertIn(expected_content, r.read())
        else:
            self.assertEqual(r.read(), expected_content)

        return r

    def test_flow_urlopen(self):
        """Verify that cassette can read a file when using urlopen."""
        self.check_read_from_file_flow(TEST_URL, "hello world")

    def test_flow_httplib(self):
        """Verify that cassette can read a file when using httplib."""

        with cassette.play(RESPONSES_FILENAME):
            conn = httplib.HTTPConnection("127.0.0.1", 5000)
            conn.request("GET", "/index")
            r = conn.getresponse()

        self.assertEqual(r.status, 200)
        self.assertEqual(r.reason, "OK")
        self.assertEqual(r.read(), "hello world")
        self.assertEqual(self.had_response.called, True)

    def test_read_twice(self):
        """Verify that response are not empty."""

        url = TEST_URL
        expected_content = "hello world"

        with cassette.play(RESPONSES_FILENAME):
            r = urllib2.urlopen(url)
            self.assertEqual(r.read(), expected_content)
            r = urllib2.urlopen(url)
            self.assertEqual(r.read(), expected_content)

    def test_read_file_https(self):
        """Verify that cassette can read a file for an HTTPS request."""
        self.check_read_from_file_flow(TEST_URL_HTTPS, "origin",
                                       allow_incomplete_match=True)

    def test_redirects(self):
        """Verify that cassette can handle a redirect."""
        self.check_read_from_file_flow(TEST_URL_REDIRECT, "hello world redirected")

    def test_non_ascii_content(self):
        """Verify that cassette can handle non-ascii content."""

        self.check_read_from_file_flow(
            url="http://127.0.0.1:5000/non-ascii-content",
            expected_content=u"Le Mexicain l'avait achetée en viager "
            u"à un procureur à la retraite. Après trois mois, "
            u"l'accident bête. Une affaire.")





class MockedResponse(object):

    def to_dict(self):
        """Return dict representation."""

        if not hasattr(self, "attrs"):
            raise AttributeError("You need to have an 'attrs' class attr.")

        return {k: getattr(self, k) for k in self.attrs}

    @classmethod
    def from_response(self, response):
        raise NotImplementedError

    def rewind(self):
        """Make sure the file description is at the start."""
        pass



# This rather useless module is needed to prevent cyclical imports
from __future__ import absolute_import
import contextlib


def get_unpatched_urlopen():
    """Return the unpatched urlopen."""
    from cassette.patcher import unpatched_urlopen
    return unpatched_urlopen


@contextlib.contextmanager
def unpatched_httplib_context(cassette_library):
    """Create a context in which httplib is unpatched."""
    from cassette.patcher import patch, unpatch

    unpatch()
    yield
    patch(cassette_library)



import logging

from cassette import unpatched

log = logging.getLogger("cassette")


class CassetteURLOpen(object):

    def __init__(self, cassette_library):
        self._cassette_library = cassette_library

    def __call__(self, url, data=None, timeout=None):
        """Fake urlopen function."""

        lib = self._cassette_library
        cassette_name = lib.cassette_name_for_urlopen(url, data)

        if cassette_name in lib:
            response = lib[cassette_name]
            response = response.rewind()

        else:
            log.warning("Making external HTTP request: %s" % cassette_name)
            # We need to unpatch, otherwise we would be using our mocked
            # httplib.
            with unpatched.unpatched_httplib_context(lib):
                response = lib.add_response(
                    cassette_name,
                    unpatched.get_unpatched_urlopen()(url, data, timeout))

        return response



import logging
import socket
from httplib import HTTPConnection, HTTPSConnection

log = logging.getLogger("cassette")


class CassetteConnectionMixin(object):

    def request(self, method, url, body=None, headers={}):
        """Send HTTP request."""

        lib = self._cassette_library
        self._cassette_name = lib.cassette_name_for_httplib_connection(
            host=self.host,
            port=self.port,
            method=method,
            url=url,
            body=body
        )
        if self._cassette_name in lib:
            self._response = lib[self._cassette_name]
            return

        log.warning("Making external HTTP request: %s" % self._cassette_name)
        self._baseclass.request(self, method, url, body, headers)

    def getresponse(self, buffering=False):
        """Return HTTP response."""

        if buffering:
            raise NotImplemented("buffering not supported by cassette.")

        if hasattr(self, "_response"):
            return self._response

        lib = self._cassette_library
        response = HTTPConnection.getresponse(self)

        # If we were just returning the response here, the file
        # descriptor would be at the end of the file, and read() would
        # return nothing.
        response = lib.add_response(self._cassette_name, response)

        return response


class CassetteHTTPConnection(CassetteConnectionMixin, HTTPConnection):

    _baseclass = HTTPConnection

    def __init__(self, *args, **kwargs):
        HTTPConnection.__init__(self, *args, **kwargs)


class CassetteHTTPSConnection(CassetteConnectionMixin, HTTPSConnection):

    _baseclass = HTTPSConnection

    def __init__(self, host, port=None, key_file=None, cert_file=None,
                 strict=None, timeout=socket._GLOBAL_DEFAULT_TIMEOUT,
                 source_address=None):
        # Directly taken from httplib.
        HTTPConnection.__init__(self, host, port, strict, timeout,
                                source_address)
        self.key_file = key_file
        self.cert_file = cert_file



from io import StringIO
from httplib import HTTPMessage

from cassette.mocked_response import MockedResponse


class MockedHTTPResponse(MockedResponse):

    attrs = ("headers", "content", "status", "reason", "raw_headers")

    @classmethod
    def from_response(cls, response):
        """Create object from true response."""

        d = {
            "headers": dict(response.getheaders()),
            "content": response.read(),
            "status": response.status,
            "reason": response.reason,
            "raw_headers": response.msg.headers,
        }
        return cls.from_dict(d)

    @classmethod
    def from_dict(cls, data):
        """Create object from dict."""

        obj = cls()

        for k in cls.attrs:
            setattr(obj, k, data[k])

        # obj.content = obj.content.decode("utf-8")
        obj.fp = StringIO(unicode(obj.content, "utf-8"))

        obj.msg = HTTPMessage(StringIO(unicode()), 0)
        for k, v in obj.headers.iteritems():
            obj.msg.addheader(k, v)

        obj.msg.headers = data["raw_headers"]

        return obj

    def read(self, chunked=None):
        return self.fp.read()

    def getheaders(self):
        return self.headers.items()

    def getheader(self, name):
        return self.headers[name]

    def close(self):
        pass



from __future__ import absolute_import
import contextlib
import logging

from cassette.cassette_library import CassetteLibrary
from cassette.patcher import patch, unpatch

cassette_library = None
logging.getLogger("cassette").addHandler(logging.NullHandler())


def insert(filename):
    """Setup cassette.

    :param filename: path to .yaml where requests and responses will be stored.
    """
    global cassette_library

    cassette_library = CassetteLibrary(filename)
    patch(cassette_library)


def eject():
    """Remove cassette, unpatching HTTP requests."""
    unpatch()


@contextlib.contextmanager
def play(filename):
    """Use cassette."""
    insert(filename)
    yield
    eject()



import urllib
import io

from cassette.mocked_response import MockedResponse


class MockedAddInfoURL(urllib.addinfourl, MockedResponse):
    """Subclass of this crazy addinfourl object that is defined in urllib."""

    attrs = ("url", "headers", "code", "content")

    @classmethod
    def create_fake(cls, url, headers, code, content):
        """Create a fake Response.

        This class method is needed so that we don't have to change the default
        constructor signature.
        """

        fp = cls.create_file_descriptor(content)

        # In urllib, constructor defined as
        # __init__(self, fp, headers, url, code=None)
        obj = cls(fp, headers, url, code)
        obj.content = content
        obj.read = fp.read

        return obj

    @classmethod
    def from_response(cls, info_url):
        """Create from an info_url object (returned by urllib2.urlopen)."""
        return cls.from_dict({
            "url": info_url.url,
            "headers": info_url.headers,
            "code": info_url.code,
            "content": info_url.read()
        })

    @classmethod
    def from_dict(cls, data):
        """Create from dict representation.

        :param dict data:
        """
        return cls.create_fake(**data)

    @staticmethod
    def create_file_descriptor(content):

        try:
            fp = io.StringIO(unicode(content))
        except UnicodeDecodeError:
            # FIXME: far too naive
            # Are you sure it's not a binary file?
            fp = io.StringIO(unicode(content, "utf-8"))

        return fp

    def rewind(self):
        self.fp = self.create_file_descriptor(self.content)
        self.read = self.fp.read
        return self



#!/usr/bin/env python
from setuptools import setup, find_packages

__version__ = "0.1.10"


def read_long_description(filename="README.rst"):
    with open(filename) as f:
        return f.read().strip()


def read_requirements(filename="requirements.txt"):
    with open(filename) as f:
        return f.readlines()

setup(
    name="cassette",
    version=__version__,
    author="Charles-Axel Dein",
    author_email="charles@uber.com",
    url="https://github.com/uber/cassette",
    license="MIT",
    packages=find_packages(),
    keywords=["http", "tests", "mock"],
    description="Cassette is a testing tool that stores external HTTP request in YAML file.",
    long_description=read_long_description(),
    install_requires=read_requirements(),
    zip_safe=False,
    test_suite="nose.collector",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development :: Testing",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ]
)
